#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: mojo spec (branch & stage) watcher."""

import argparse
import logging
import os
import sys
import subprocess


from ci_automation.branch import (
    get_info_from_revno_dict,
    get_info_message_from_revno_dict,
    get_revnos_for_spec,
)
from ci_automation.mojo_project_pool import (
    MojoProjectPool,
    NoProjectError,
)


MOJO_PROJECT_POOL_DIR = os.path.expanduser('~/mojo_project_pool')
MOJO_PROJECT_POOL_LOCK_DIR = os.path.expanduser('~/mojo_project_pool/locks')


def deploy(args):
    """Call the mojo wrapper script with the given arguments.

    Swallow any failures.
    """
    mojo_cmd = os.path.join(os.path.dirname(__file__), 'mojo.py')
    return subprocess.call([mojo_cmd] + args)


def main():

    parser = argparse.ArgumentParser(
        description=('Deploys mojo spec as a new environment if it is not '
                     'yet deployed.')
    )
    parser.add_argument('--branch', required=True)
    parser.add_argument('--stage', required=True)
    parser.add_argument('--use-project-pool',
                        action='store_true', default=False)

    # Parse known arguments, collect (and augment) the rest as 'mojo_args'.
    args, mojo_args = parser.parse_known_args()
    if args.use_project_pool and '--project' in mojo_args:
        print("Error: Cannot specify both --use-project-pool and --project")
        return 1

    mojo_args.extend([
        '--branch', args.branch,
        '--stage', args.stage
    ])

    logging.basicConfig(
        format='%(asctime)s  %(name)s %(levelname)s: %(message)s')
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Optional remote logging via logstash.
    logstash_host = os.environ.get('CI_LOGSTASH_HOST')
    if logstash_host is not None:
        try:
            import logstash
        except ImportError:
            print('Follow the README instructions for installing '
                  'python-logstash')
            return 1
        else:
            logger.addHandler(
                logstash.LogstashHandler(logstash_host, 5959, 1))

    # Deploy new environment if the stage identifier isn't recorded.
    idenfifier_dir = os.path.expanduser('~/ci-cd-identifiers/')
    if not os.path.exists(idenfifier_dir):
        os.makedirs(idenfifier_dir)

    spec_revno_dict = get_revnos_for_spec(args.branch, args.stage)
    identifier = get_info_from_revno_dict(spec_revno_dict)
    info = get_info_message_from_revno_dict(spec_revno_dict)

    identifier_path = os.path.join(idenfifier_dir, identifier)
    if os.path.exists(identifier_path):
        return 0

    # Thomi - 2015-05-28: Transitional code to use the mojo project pool code.
    # Once this is in production we can make this the default. Until then,
    # this giant if statement determines if we use the pool or a dedicated
    # project from the command line:
    if args.use_project_pool:
        # use the project pool:
        pool = MojoProjectPool(
            MOJO_PROJECT_POOL_DIR,
            MOJO_PROJECT_POOL_LOCK_DIR
        )
        try:
            with pool.acquire_mojo_project() as project_name:
                # Register deployment identifier regardless its deployment
                # status. We don't want to deploy it again.
                with open(identifier_path, 'w') as fd:
                    fd.write(info + '\n')

                mojo_args += ['--project', project_name]
                status = 'SUCCEEDED' if deploy(mojo_args) == 0 else 'FAILED'
        except NoProjectError:
            print("No available mojo projects, not deploying.")
            print("Deployment will be retried next run.")
            return 0
    else:
        # Don't use the project pool - use whatever was supplied on the CLI:
        # Register deployment identifier regardless its deployment status.
        # We don't want to deploy it again.
        with open(identifier_path, 'w') as fd:
            fd.write(info + '\n')
        # `mojo.py` prints additional information to stdout, which is only
        # recorded locally, for now ...
        status = 'SUCCEEDED' if deploy(mojo_args) == 0 else 'FAILED'

    # We only log a sucint but descriptive message about the deployment
    # status. This content will be shipped to logstash and be handled
    # from there (recording, stats, pagerduty, etc).
    extra = {
        'service': 'ci/cd',
        'deployment_status': status,
        'deployment_branch': args.branch,
        'deployment_name': os.path.dirname(args.stage),
        'deployment_identifier': identifier,
        'deployment_info': info,
    }
    logger.info(
        'Deploy attempt for %s %s:\n====\n%s\n====',
        identifier, status, info, extra=extra)

    # Never fail as far as cron is concerned.
    return 0


if __name__ == '__main__':
    sys.exit(main())
