#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation branch helpers."""

from collections import OrderedDict
import hashlib
import os
import subprocess
import re


from . import config


def get_revno(url):
    '''Get the current revision for a branch URL.

    If the URL contains a fixed revno, return it.'''

    match = re.search(';revno=([0-9]*)', url)
    if match:
        # The revno is fixed. Return it.
        return match.group(1)

    out = subprocess.check_output(['bzr', 'revno', url],
                                  stderr=subprocess.DEVNULL)
    return out.decode('utf8').strip('\n')


def get_path_revno(url, path):
    '''Return the last change revision of the path in the branch URL.'''

    if ';revno=' in url:
        # The revno is fixed. Return it.
        return get_revno(url)

    p = '{}/{}'.format(url, path)
    out = subprocess.check_output(['bzr', 'log', '-S', '-l', '1', p],
                                  stderr=subprocess.DEVNULL)
    out = out.decode('utf8')
    match = re.search(' *([0-9]*)', out)
    if match:
        return match.group(1)
    else:
        return None


def file_contents_on_branch(branch, stage, path):
    '''Get the contents of the given file at the given branch.

    Returns empty-string if the file does not exist and cannot be accessed.
    '''

    if not branch.endswith('/'):
        branch += '/'
    full_path = branch + os.path.join(stage, path)
    try:
        out = subprocess.check_output(['bzr', 'cat', full_path],
                                      stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return ''
    return out.decode('utf8').strip('\n')


def branches_in_contents(contents):
    '''Parse out the branches contained in the given file and return a
    generator of them.'''

    for line in contents.split('\n'):
        # Ignore comments
        if line.startswith('#'):
            continue
        match = re.search('\s((?:lp:|bzr\+ssh:|http:)\S*)', line)
        if match:
            yield match.group(1)


def get_revnos_for_spec(spec_branch, stage):
    """Get bzr revnos for a spec branch and it's contents.

    'spec_branch' should be a valid uri to a mojo spec in bzr.
    'stage' should be the name of a stage within the mojo spec.

    This returns an ordered dictionary. Keys are the bzr branch names, values
    are the bzr revnos for that branch.

    If the spec has configuration on disk, the dictionary will include
    a bzr revno for the 'configuration' key.

    """
    identifiers = OrderedDict()

    # Cover changes in stage (name & revno)
    stage_revno = get_path_revno(spec_branch, os.path.dirname(stage))
    identifiers[stage] = stage_revno

    # Cover changes in collected branches (url/path and revno),
    # considers stage-specific and top-level 'collect' files.
    for stage in (stage, os.path.dirname(stage)):
        contents = file_contents_on_branch(spec_branch, stage, 'collect')
        for b in branches_in_contents(contents):
            identifiers[b] = get_revno(b)

    if config.deployment_has_config(stage):
        identifiers['configuration'] = config.get_revno_for_deployment(stage)
    return identifiers


def get_info_from_revno_dict(revno_dict):
    """Get a shortened sha1 digest given a revno_dict, as produced by
    get_revnos_for_spec.

    """
    identifiers = ["{}\t{}".format(k, v) for k, v in revno_dict.items()]
    info = '\n'.join(identifiers)
    return hashlib.sha1(info.encode('utf-8')).hexdigest()[-8:]


def get_info_message_from_revno_dict(revno_dict):
    """Get a string that contains all the revisions we deployed.

    Unlike get_info_from_revno_dict, this function can change over time, since
    it's just used for human consumption.

    """
    identifiers = ["{}\t{}".format(k, v) for k, v in revno_dict.items()]
    return '\n'.join(identifiers)
