#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation config helpers.

We store config files on disk as fully realised files (no templating nonsense)
and copy them to the mojo "secrets" folder before calling mojo to deploy the
service in question.

This gives us several advantages:

1. It's super-simple to see what config a service will get deployed with -
   just look in the config directory and examine the last bzr commit contents.

2. We get a log of all changes, who made them, and *why*, since we only get
   the last bzr commit - unstaged changes are never deployed.

3. We can re-deploy a service by changing the config file, since the config
   folder's bzr revno is considered to be part of the deployment hash.

All configs must be stored in `CONFIG_DIR`. By default this is set to
~/service_configs. All functions read this variable without caching, so tests
can be written by setting this variable to something else for the lifetime of
the test.

Beneath the config_dir, the folder structure should mimic the
canonical-mojo-specs project. For example::

    service_configs
     - ue
      - mojo-ue-adt-cloud-service
       - devel
       - production
      - mojo-ue-adt-cloud-worker
       - devel
       - production

...continued for every mojo spec we want to deploy. Within every stage
directory ('devel' / 'production' etc), any files present will be present in
the mojo secrets dir when deployment begins.

"""

import os.path
import re
import shutil
import subprocess


__all__ = [
    'check_config_dir_exists',
    'deployment_has_config',
    'export_latest_config_for_deployment',
    'get_revno_for_deployment',
]


CONFIG_DIR = os.path.expanduser("~/service_configs")


def check_config_dir_exists():
    """Return True if the config directory exists, and is a bzr repository."""
    global CONFIG_DIR
    if not os.path.exists(CONFIG_DIR):
        return False
    return subprocess.call(
        ['bzr', 'info'],
        cwd=CONFIG_DIR,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    ) == 0


def deployment_has_config(mojo_stage):
    """Return true if a mojo spec stage has configuration data on disk.

    'mojo_stage' should be the name of a mojo stage to deploy, like
    'ue/mojo-ue-adt-cloud-service/devel' or
    'ue/mojo-ue-adt-cloud-service/production'.

    """
    global CONFIG_DIR
    full_path = os.path.join(
        CONFIG_DIR,
        mojo_stage,
    )
    return (
        os.path.exists(full_path) and
        os.path.isdir(full_path) and
        get_revno_for_deployment(mojo_stage) is not None
    )


def get_revno_for_deployment(mojo_stage):
    """Get the bzr revno for the given mojo spec & stage.

    'mojo_stage' should be the name of a mojo stage to deploy, like
    'ue/mojo-ue-adt-cloud-service/devel' or
    'ue/mojo-ue-adt-cloud-service/production'.

    """
    global CONFIG_DIR
    try:
        output = subprocess.check_output(
            ['bzr', 'log', '-l', '1', '-S', mojo_stage],
            cwd=CONFIG_DIR,
            stderr=subprocess.DEVNULL,
        ).decode()
    except subprocess.CalledProcessError as e:
        # returncode 3 is 'unknown path'
        if e.returncode != 3:
            raise
        return None
    match = re.search(' *([0-9]+)', output)
    if match:
        return match.group(1)
    return None


def export_latest_config_for_deployment(mojo_stage, target_dir):
    """Export configuration for a mojo spec to 'target_dir'.

    'mojo_stage' should be the name of a mojo stage to deploy, like
    'ue/mojo-ue-adt-cloud-service/devel' or
    'ue/mojo-ue-adt-cloud-service/production'.

    'target_dir' is the target location to export to. This will probably be
    set to '/srv/mojo/LOCAL' in production.

    Directories for the spec and stage will be created in the target directory.

    :raises: ValueError if the given mojo spec & stage don't have configuration
        in the config dir.

    """
    global CONFIG_DIR
    if not deployment_has_config(mojo_stage):
        raise ValueError(
            "Mojo stage '{}' have no config on disk.".format(mojo_stage)
        )
    target_dir = os.path.join(target_dir, mojo_stage)
    config_dir = os.path.join(CONFIG_DIR, mojo_stage)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    else:
        for entry in os.listdir(target_dir):
            full_path = os.path.join(target_dir, entry)
            if os.path.isdir(full_path):
                shutil.rmtree(full_path)
            else:
                os.unlink(full_path)
    subprocess.check_call(
        ['bzr', 'export', target_dir, config_dir],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
