# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: juju helpers"""

import os
import shutil
import subprocess
import time
import yaml

from . import utils
from . import nova


def deployment_has_floating_ips(name, base):
    """Returns True if the juju deploymet has any floating ip addresses."""
    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir

    args = ['juju', 'status']
    juju_status = yaml.load(
        subprocess.check_output(args, env=env)
    )
    nova_instance_ids = [
        m['instance-id'] for m in juju_status['machines'].values()
    ]
    return any([nova.instance_has_floating_ip(i) for i in nova_instance_ids])


def destroy_all_services(name, base):
    """Destroy all services for a particular juju environment.

    This is necessary since destroying the environment does not shut down
    services cleanly. This can result in (for example) stale rabbit
    connections.
    """
    # This can take a while, so let's limit the damage to 5 minutes:
    timeout = 60 * 5
    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir
    try:
        for service in get_active_services(name, base):
            subprocess.call(
                ['juju', 'destroy-service', service],
                env=env
            )
        start_time = time.time()
        while time.time() - start_time < timeout:
            services = get_active_services(name, base)
            if len(services) == 0:
                return
            time.sleep(29)
    except subprocess.CalledProcessError as e:
        print("juju status failed: %s" % e)
        return


def get_active_services(name, base):
    """Get a list of the services currently active.

    Ignores ksplice, landscape and basenode.

    """
    ignored_services = (
        'ksplice',
        'landscape-client',
        'ubuntu-basenode',
    )
    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir

    args = ['juju', 'status']
    juju_status = yaml.load(
        subprocess.check_output(args, env=env)
    )
    return [s for s in juju_status['services'] if s not in ignored_services]


def destroy_environment(name, base):
    # We can safely destroy the existing environment, as the services are
    # promised to be idempotent.

    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir

    try:
        args = ['juju', 'status']
        subprocess.check_call(
            args, env=env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
    except subprocess.CalledProcessError:
        # No juju environment here.
        pass
    else:
        args = ['juju', 'destroy-environment', '-y', '--force', name]
        try:
            utils.check_call(args, env=env)
        except subprocess.CalledProcessError as e:
            print("juju destroy-environment failed with %s" % e)
            print("Manually deleting left-over instances...")
            # juju sometimes fails, so simulate what it does ourselves.
            remaining = nova.get_instances_for_environment(name)
            nova.delete_instances(remaining)
            print("Instances have been manually deleted")

    # Ensure all related security-groups left behind by `juju` are removed,
    # otherwise they will remain consuming resources (and quota).
    nova_output = subprocess.check_output(['nova', 'secgroup-list'])
    secgroups = [
        line.split('|')[2].strip()
        for line in nova_output.decode('utf-8').splitlines()
        if name in line
    ]

    # Try harder because sometimes the instance retaining the secgroup
    # takes sometime to be purged.
    for i in range(3):
        for secgroup in secgroups.copy():
            try:
                subprocess.check_call(
                    ['nova', 'secgroup-delete', secgroup],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
                secgroups.remove(secgroup)
            except subprocess.CalledProcessError:
                print(
                    "Error removing secgroup '{}'. "
                    "It will be retried".format(secgroup)
                )
        time.sleep(1)

    if secgroups:
        print(
            'The security groups "{}" could not be removed. '
            'Please remove them manually.'.format(', '.join(secgroups))
        )

    shutil.rmtree(env_dir)


def new_environment(name, base, devel=False, network=None):
    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir

    if os.path.exists(env_dir):
        destroy_environment(name, base)

    os.makedirs(env_dir)
    contents = {
        'default': name,
        'environments': {
            name: {
                'type': 'openstack',
                'control-bucket': 'juju-{}'.format(name),
                # If we ever remove this, we'll need to manually
                # set security group rules allowing ssh from
                # 91.189.90.53/32. Juju creates these rules for
                # 0.0.0.0, which juju-reaper removes (would
                # allow public SSH for any instance with a
                # floating IP).
                'use-default-secgroup': 'true',
            },
        },
    }

    if network:
        contents['environments'][name]['network'] = network

    if not devel:
        # Otherwise we chew up Prodstack's disk IO
        # upgrading, but it doesn't work for 'bootstack' (our devel).
        contents['environments'][name]['image-stream'] = 'daily'

    with open(os.path.join(env_dir, 'environments.yaml'), 'w') as fp:
        fp.write(yaml.dump(contents, default_flow_style=False))

    args = ['juju', 'bootstrap']
    utils.check_call(args, env=env)

    return env_dir


def set_metadata(name, base, metadata):
    """Set deployment metadata.

    'name' is the juju environment name.
    'base' is the juju environment directory.
    'metadata' is a mapping of metadata you want to set.

    This function will never raise an exception - if setting metadata fails,
    an error will be printed, but that will never bubble up out of this
    function, so deployments won't be compromised.

    """
    env_dir = os.path.join(base, name)
    env = os.environ.copy()
    env['JUJU_HOME'] = env_dir

    try:
        args = ['juju', 'status']
        juju_status = subprocess.check_output(
            args, env=env,
            stderr=subprocess.DEVNULL
        ).decode()
    except Exception as e:
        print(
            "Could not set deployment metadata. "
            "'juju stat' with JUJU_HOME=%s failed."
            "This is a non-fatal error: %s" % (env_dir, e)
        )
        return
    try:
        data = yaml.load(juju_status)
        for juju_machine_id in data['machines'].keys():
            nova_instance = data['machines'][juju_machine_id]['instance-id']
            nova.set_metadata_on_instance(nova_instance, metadata)
    except Exception as e:
        print("Unable to set nova metadata: %s" % e)
