# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: mojo project pool.

Instead of maintaining a 1:1 mapping between a mojo project and a service it
deploys, let's use a pool of mojo projects.

The code in this file helps us do that while avoiding race conditions.

The MojoProjectPool class handles allocating and deallocating mojo projects.
It does this with a directory on disk. This directory is called the 'pool
directory' in documentation.

The pool directory must contain files, the names of which are each mojo
project name. These files must be created manually. If a mojo project is
represented in the pool directory, then it is used exclusively by the
deployer, and MUST NOT be used manually (since manual invocation of mojo
will not update lock files).

MojoProjectPool allocates projects by creating symlinks on disk to project
files. We use symlinks because:
- symlink creation is atomic
- symlink counts can be read from the lockfile inode.

Symlinks are created in the 'project lock directory'.
"""

from contextlib import contextmanager
import os
import os.path
import random


class NoProjectError(Exception):

    def __str__(self):
        return "No mojo projects available"


class MojoProjectPool(object):

    def __init__(self, project_dir, lock_dir):
        if not os.path.exists(project_dir):
            os.makedirs(project_dir)
        self.project_dir = project_dir
        if not os.path.exists(lock_dir):
            os.makedirs(lock_dir)
        self.lock_dir = lock_dir

    @contextmanager
    def acquire_mojo_project(self):
        """Get a mojo project to use.

        This contextmanager yields the mojo project name, so it can be used
        like so:

        with project_pool.acquire_mojo_project() as mojo_project_name:
            print("Using mojo project: {}".format(project_name))

        If no project can be found, a NoProjectError exception is raised.

        """
        candidates = [
            f for f in os.listdir(self.project_dir)
            if os.path.isfile(os.path.join(self.project_dir, f))
        ]
        random.shuffle(candidates)
        for candidate_project in candidates:
            try:
                os.symlink(
                    candidate_project,
                    os.path.join(self.lock_dir, candidate_project)
                )
            except FileExistsError:
                continue
            # If we get here, the lock is acquired:
            try:
                yield candidate_project
            finally:
                os.unlink(os.path.join(self.lock_dir, candidate_project))
                return
        else:
            raise NoProjectError()
