# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: nova helpers."""

import io
import os
import stat
import subprocess

from . import utils


def nova_list():
    """Default 'nova list' accessor."""
    args = [
        'nova',
        'list',
        '--fields', 'name,networks',
    ]

    output = subprocess.check_output(
        args, stderr=subprocess.DEVNULL).decode('utf-8').strip()

    return output


def get_list_of_nova_instances(nova_output=None):
    """A generator that yields all the current nova instances."""
    if nova_output is None:
        nova_output = nova_list()

    for line in io.StringIO(nova_output).readlines()[3:-1]:
        yield line.split('|')[2].strip()


def get_instances_for_environment(env_name):
    """get a list of all instnaces whose names start with 'env_name'.

    This is a hacky way of getting all the nova instances that are part of
    a given juju environment.

    """
    matched = []
    for instance_name in get_list_of_nova_instances():
        if instance_name.strip().startswith(env_name):
            matched.append(instance_name)
    return matched


def get_services_map(base_dir, nova_output=None):
    """Return a services map dictionary.

    'base_dir' is the path where deployment identifier can be found.

    If 'nova_output' is not provided `nova_list` accessor is used to get
    this information.

    Returns a dictionary key-ed by service name (middle term of the mojo
    stage used in the deployment, e.g. 'mojo-ue-core-image-watcher'):

    {
        'mojo-ue-core-image-watcher': {
            'DEADBEEF': {
                'path': <identifier_path>,
                'epoch': <identifier_ST_MTIME>,
                'units': [<unit_name>, <unit_name>, ...],
            },
            'ABCD1234': {
                'path': <identifier_path>,
                'epoch': <identifier_ST_MTIME>,
                'units': [<unit_name>, <unit_name>, ...],
            },
        },
    }

    Manually deployed environments, i.e. no identifiers on disk, are ignored.
    """
    if nova_output is None:
        nova_output = nova_list()

    services = {}

    for line in io.StringIO(nova_output).readlines()[3:-1]:
        unit_name = line.split('|')[2].strip()
        service_name = '-'.join(unit_name.split('-')[1:-3])
        try:
            identifier = unit_name.split('-')[-3]
        except IndexError:
            continue

        path = os.path.join(base_dir, identifier)
        if not os.path.exists(path):
            continue

        service = services.setdefault(service_name, {})
        deployment = service.setdefault(identifier, {})
        units = deployment.setdefault('units', [])
        units.append(unit_name)

        if deployment.get('path') is None:
            deployment.update({
                'path': path,
                'epoch': os.stat(path)[stat.ST_MTIME],
            })

        deployment.setdefault('floating_ips', [])
        networks = line.split('|')[3].strip()
        if networks:
            ips = networks.split('=')[1].strip()
            floating_ips = [ip.strip() for ip in ips.split(',')][1:]
            deployment['floating_ips'].extend(floating_ips)

    return services


def set_metadata_on_instance(instance_name, metadata):
    """Set some metadata on a nova instance.

    :param instance_name: The nova instance name or id. If using the instance
        name, it is the callers responsibility to ensure that the instance
        name is unique.

    :param metadata: a mapping of key/value pairs to set as metadata. keys
        cannot contain '=' characters, but may contain spaces. Values may
        contain any valid unicode.

    """
    meta_parameters = []
    for key, value in metadata.items():
        if '=' in key:
            raise ValueError("metadata Keys may not contain '='")
        key = key.replace('"', '\\"')
        if ' ' in key:
            key = '"{}"'.format(key)
        value = value.replace('"', '\\"')
        if ' ' in value:
            value = '"{}"'.format(value)
        meta_parameters.append('{}={}'.format(key, value))
    if meta_parameters:
        subprocess.check_call(
            ['nova', 'meta', instance_name, 'set'] + meta_parameters
        )


def instance_has_floating_ip(instance_name):
    """
    Return true if the given nova instance has a floating IP address
    associated with it.
    """
    output_lines = subprocess.check_output(
        ['nova', 'show', instance_name]
    ).decode().split('\n')[3:-1]
    for line in output_lines:
        key, value = [l.strip() for l in line.split('|') if l]
        if 'network' in key:
            return len(value.split(',')) > 1
    # If there is no network key (unlikely), treat it as not having a floating
    # ip:
    return False


def delete_instances(instance_list):
    for i in instance_list:
        delete_instance(i)


def delete_instance(instance_name):
    utils.check_call(['nova', 'delete', instance_name])
