#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation tests for branch helpers."""

from collections import OrderedDict
import os
import shutil
import subprocess
import sys
import tempfile
import unittest


HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, '../..'))

from ci_automation.branch import (
    get_info_from_revno_dict,
    get_path_revno,
    get_revno,
    get_revnos_for_spec,
)


class TestBranchUtils(unittest.TestCase):

    def setUp(self):
        super(TestBranchUtils, self).setUp()
        self.branch = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.branch)
        self.branch_init()

    def branch_init(self):
        subprocess.check_call(['bzr', 'init'], cwd=self.branch,
                              stderr=subprocess.DEVNULL,
                              stdout=subprocess.DEVNULL)

    def branch_add(self, path):
        subprocess.check_call(['bzr', 'add', path],
                              cwd=self.branch,
                              stderr=subprocess.DEVNULL,
                              stdout=subprocess.DEVNULL)

    def branch_commit(self, message='Bazinga!'):
        subprocess.check_call(['bzr', 'commit', '-m', message],
                              cwd=self.branch,
                              stderr=subprocess.DEVNULL,
                              stdout=subprocess.DEVNULL)

    def branch_add_stage(self, name):
        stage = os.path.join(self.branch, name)
        os.makedirs(stage)
        with open(os.path.join(stage, 'collect'), 'w') as fd:
            fd.write('foo lp:foo;revno=1')
        self.branch_add(stage)
        self.branch_commit()

    def branch_update_collect(self, stage, content='foo lp:foo;revno=1'):
        collect = os.path.join(self.branch, stage, 'collect')
        with open(collect, 'w') as fd:
            fd.write(content)
        self.branch_commit()

    def test_get_revno(self):
        self.assertEqual(get_revno(self.branch), '0')

    def test_get_revno_with_embedded(self):
        self.assertEqual(get_revno('lp:fake;revno=33'), '33')

    def test_get_revno_with_path(self):
        paths = (os.path.join(self.branch, 'bob'),
                 os.path.join(self.branch, 'bob2'))
        for path in paths:
            with open(path, 'w'):
                pass
            self.branch_add(path)
            self.branch_commit()
        self.assertEqual('1', get_path_revno(self.branch, 'bob'))
        self.assertEqual('2', get_path_revno(self.branch, 'bob2'))

    def test_get_revnos_for_spec(self):
        stage = 'bob/devel'
        expected = {
            'bob/devel': '1',
            'lp:foo;revno=1': '1'
        }
        self.branch_add_stage(stage)
        self.assertEqual(
            expected,
            get_revnos_for_spec(self.branch, stage),
        )

    def test_get_revnos_for_spec_ignores_other_stages(self):
        self.branch_add_stage('bob/devel')
        self.branch_add_stage('bob/production')
        self.branch_update_collect('bob/production', 'test lp:test;revno=3')

        devel_specs = get_revnos_for_spec(self.branch, 'bob/devel')
        prod_specs = get_revnos_for_spec(self.branch, 'bob/production')

        self.assertEqual('1', devel_specs['lp:foo;revno=1'])
        self.assertEqual('3', prod_specs['lp:test;revno=3'])

    def test_get_info_from_revno_dict_gives_8_char_hash(self):
        spec_dict = OrderedDict()
        spec_dict['bob/devel'] = '234'
        spec_dict['lp:foo'] = '134'
        sha = get_info_from_revno_dict(spec_dict)
        self.assertEqual(8, len(sha))

    def test_get_info_from_revno_dict_produces_new_hash_with_new_dict(self):
        spec_dict = OrderedDict()
        spec_dict['bob/devel'] = '234'
        spec_dict['lp:foo'] = '134'
        sha1 = get_info_from_revno_dict(spec_dict)
        spec_dict['lp:foo'] = '135'
        sha2 = get_info_from_revno_dict(spec_dict)
        self.assertNotEqual(sha1, sha2)


if __name__ == '__main__':
    unittest.main()
