#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation tests for config helpers."""

import tempfile
import unittest
import shutil
import functools
import contextlib
import subprocess
import os.path

from ci_automation import config


class TestConfig(unittest.TestCase):

    def test_config_dir_check_returns_false_if_dir_does_not_exist(self):
        with tempfile.TemporaryDirectory() as workdir:
            config_dir = os.path.join(workdir, "doesnotexist")
            with patch_config_dir(config_dir):
                self.assertEqual(False, config.check_config_dir_exists())

    def test_config_dir_check_returns_false_if_not_a_bzr_repo(self):
        with tempfile.TemporaryDirectory() as config_dir:
            with patch_config_dir(config_dir):
                self.assertEqual(False, config.check_config_dir_exists())

    def test_config_dir_returns_true_on_empty_repo(self):
        with BzrConfigDir() as cm:
            self.assertEqual(True, config.check_config_dir_exists())

    def test_deployment_has_config_returns_false_on_nonexisting_project(self):
        with BzrConfigDir() as cm:
            self.assertEqual(
                False,
                config.deployment_has_config('ue/nonexist/devel'),
            )

    def test_deployment_has_config_returns_false_on_nonexisting_stage(self):
        with BzrConfigDir() as cm:
            os.makedirs(os.path.join(cm.config_dir, 'ue/test'))
            self.assertEqual(
                False,
                config.deployment_has_config('ue/test/devel'),
            )

    def test_deployment_has_config_returns_true_on_existing_stage(self):
        with BzrConfigDir() as cm:
            cm.add_mojo_stage('ue/my-test-project/devel')
            self.assertEqual(
                True,
                config.deployment_has_config('ue/my-test-project/devel'),
            )

    def test_get_revno_for_deployment_returns_none_without_bzr_control(self):
        with BzrConfigDir() as cm:
            os.makedirs(os.path.join(cm.config_dir, 'ue/project/devel'))
            self.assertIsNone(
                config.get_revno_for_deployment('ue/project/devel')
            )

    def test_get_revno_for_deployment_returns_revno(self):
        with BzrConfigDir() as cm:
            cm.add_mojo_stage('ue/project/devel')
            self.assertEqual(
                '1',
                config.get_revno_for_deployment('ue/project/devel'),
            )

    def test_adding_file_increases_bzr_revno(self):
        with BzrConfigDir() as cm:
            project = cm.add_mojo_stage('ue/project/production')
            with project.write_new_file('config.ini') as f:
                f.write('[section]\n')
                f.write('key = value\n')
            self.assertEqual(
                '2',
                config.get_revno_for_deployment('ue/project/production'),
            )

    def test_projects_dont_increase_each_others_revnos(self):
        with BzrConfigDir() as cm:
            project1 = cm.add_mojo_stage('ue/project1/production')
            project2 = cm.add_mojo_stage('ue/project2/production')
            self.assertEqual(
                '1',
                config.get_revno_for_deployment('ue/project1/production'),
            )
            self.assertEqual(
                '2',
                config.get_revno_for_deployment('ue/project2/production'),
            )
            with project1.write_new_file('README') as f:
                f.write("Some text")
            self.assertEqual(
                '3',
                config.get_revno_for_deployment('ue/project1/production'),
            )
            self.assertEqual(
                '2',
                config.get_revno_for_deployment('ue/project2/production'),
            )

    def test_export_raises_ValueError_for_unconfigured_spec(self):
        with BzrConfigDir() as cm:
            with self.assertRaises(ValueError):
                config.export_latest_config_for_deployment(
                    'project/devel',
                    'target'
                )

    def test_exports_empty_config_dir(self):
        with BzrConfigDir() as cm, tempfile.TemporaryDirectory() as target_dir:
            cm.add_mojo_stage('ue/project/production')
            target_project_path = os.path.join(
                target_dir,
                'ue/project/production'
            )
            config.export_latest_config_for_deployment(
                'ue/project/production',
                target_dir
            )
            self.assertTrue(os.path.exists(target_project_path))

    def test_exports_simple_file(self):
        with BzrConfigDir() as cm, tempfile.TemporaryDirectory() as target_dir:
            project = cm.add_mojo_stage('ue/project/production')
            with project.write_new_file('config.ini') as f:
                f.write("Test config")
            config.export_latest_config_for_deployment(
                'ue/project/production',
                target_dir
            )
            target_config_path = os.path.join(
                target_dir,
                'ue/project/production/config.ini'
            )
            self.assertTrue(os.path.exists(target_config_path))
            with open(target_config_path) as f:
                self.assertEqual(f.read(), "Test config")

    def test_exports_directory_heirarchy(self):
        with BzrConfigDir() as cm, tempfile.TemporaryDirectory() as target_dir:
            project = cm.add_mojo_stage('ue/project/production')
            project.create_directory('foo/bar/baz')
            with project.write_new_file('foo/bar/baz/config.ini') as f:
                f.write("Test config")
            config.export_latest_config_for_deployment(
                'ue/project/production',
                target_dir
            )
            target_config_path = os.path.join(
                target_dir,
                'ue/project/production/foo/bar/baz/config.ini'
            )
            self.assertTrue(os.path.exists(target_config_path))
            with open(target_config_path) as f:
                self.assertEqual(f.read(), "Test config")


@contextlib.contextmanager
def patch_config_dir(new_dir):
    """A context manager that patches config.CONFIG_DIR."""
    old_dir = config.CONFIG_DIR
    config.CONFIG_DIR = new_dir
    try:
        yield
    except:
        config.CONFIG_DIR = old_dir
        raise


class BzrConfigDir(contextlib.ExitStack):
    """A context manager that sets up a simple config bzr repo."""

    def __enter__(self):
        super().__enter__()
        self.config_dir = self.enter_context(tempfile.TemporaryDirectory())
        self.enter_context(patch_config_dir(self.config_dir))

        self._create_config_repo()
        return self

    def _create_config_repo(self):
        subprocess.check_call(
            ['bzr', 'init'],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=self.config_dir,
        )

    def add_mojo_stage(self, spec_path):
        full_path = os.path.join(self.config_dir, spec_path)
        os.makedirs(full_path)
        subprocess.check_call(
            ['bzr', 'add', spec_path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=self.config_dir,
        )
        subprocess.check_call(
            ['bzr', 'commit', '-m', 'adding {}'.format(spec_path)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=self.config_dir,
        )
        return ProjectConfigManager(full_path)


class ProjectConfigManager(object):

    """A simple interface to make it easy to write data to a test config
    dir for a mojo project.
    """

    def __init__(self, path):
        self.path = path

    @contextlib.contextmanager
    def write_new_file(self, path):
        full_path = os.path.join(self.path, path)
        with open(full_path, 'wt') as f:
            yield f
        subprocess.check_call(
            ['bzr', 'add', path],
            cwd=self.path,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        subprocess.check_call(
            ['bzr', 'commit', '-m', 'adding {}'.format(path)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=self.path,
        )

    def create_directory(self, dir_path):
        full_path = os.path.join(self.path, dir_path)
        os.makedirs(full_path)
        subprocess.check_call(
            ['bzr', 'add', dir_path],
            cwd=self.path,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        subprocess.check_call(
            ['bzr', 'commit', '-m', 'adding {}'.format(dir_path)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=self.path,
        )


if __name__ == '__main__':
    unittest.main()
