#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
import tempfile
import unittest
import shutil

from ci_automation.mojo_project_pool import (
    MojoProjectPool,
    NoProjectError,
)


class MojoProjectPoolTests(unittest.TestCase):

    def test_creates_project_and_lock_directories(self):
        with tempfile.TemporaryDirectory() as tempdir:
            project_path = os.path.join(tempdir, 'projects')
            lock_path = os.path.join(tempdir, 'projects', 'locks')
            pool = MojoProjectPool(project_path, lock_path)

            self.assertTrue(os.path.exists(project_path))
            self.assertTrue(os.path.exists(lock_path))

    def test_cannot_acquire_project_when_none_are_setup(self):
        with ProjectPoolDir() as workdir:
            pool = MojoProjectPool(workdir.project_dir, workdir.lock_dir)

            with self.assertRaises(NoProjectError):
                with pool.acquire_mojo_project():
                    pass

    def test_can_acquire_single_project(self):
        with ProjectPoolDir(projects=['project-one']) as workdir:
            pool = MojoProjectPool(workdir.project_dir, workdir.lock_dir)
            with pool.acquire_mojo_project() as project:
                self.assertEqual(project, 'project-one')

    def test_can_acquire_single_project_once_only(self):
        with ProjectPoolDir(projects=['project-one']) as workdir:
            pool = MojoProjectPool(workdir.project_dir, workdir.lock_dir)
            with pool.acquire_mojo_project() as project:
                self.assertEqual(project, 'project-one')
                with self.assertRaises(NoProjectError):
                    with pool.acquire_mojo_project():
                        pass

    def test_can_acquire_multiple_projects(self):
        with ProjectPoolDir(projects=['one', 'two']) as workdir:
            pool = MojoProjectPool(workdir.project_dir, workdir.lock_dir)
            with pool.acquire_mojo_project() as p1, \
                    pool.acquire_mojo_project() as p2:
                self.assertNotEqual(p1, p2)
                self.assertIn(p1, ('one', 'two'))
                self.assertIn(p2, ('one', 'two'))

    def test_uncaught_exception_releases_lock(self):
        with ProjectPoolDir(projects=['project-one']) as workdir:
            pool = MojoProjectPool(workdir.project_dir, workdir.lock_dir)
            try:
                with pool.acquire_mojo_project() as project:
                    raise Exception("Pretending mojo failed!")
            except Exception:
                # must be able to re-acquire the project (it must have been
                # released).
                with pool.acquire_mojo_project() as project:
                    self.assertEqual(project, 'project-one')


class ProjectPoolDir(object):

    """A context manager that creates a project pool on disk."""
    def __init__(self, projects=[]):
        self.project_dir = None
        self.lock_dir = None
        self._create_projects = projects

    def __enter__(self):
        self.workdir = tempfile.mkdtemp()
        self.project_dir = os.path.join(self.workdir, 'projects')
        self.lock_dir = os.path.join(self.workdir, 'locks')
        if self._create_projects:
            os.makedirs(self.project_dir)
            for project_name in self._create_projects:
                with open(os.path.join(self.project_dir, project_name), 'w'):
                    pass
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.workdir)
        self.project_dir = None
        self.lock_dir = None


if __name__ == '__main__':
    unittest.main()
