#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation tests for nova helpers."""

import os
import shutil
import sys
import tempfile
import time
import unittest


HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, '../..'))

from ci_automation.nova import get_services_map


class TestNova(unittest.TestCase):

    identifiers = (
        '12345',
        'abcdef',
        '56789',
    )
    service_names = (
        'mojo-ue-core-result-watcher',
        'mojo-ue-core-image-watcher',
        'mojo-ue-core-image-publisher',
        'mojo-ue-core-image-tester',
    )

    def setUp(self):
        self.base_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.base_dir)

    def generate_identifiers(self):
        """Generate deployment identifiers on disk."""
        for ident in self.identifiers:
            # get unique timestamps
            time.sleep(.1)
            os.makedirs(os.path.join(self.base_dir, ident))

    def create_output(self):
        """Return some fake 'nova list' output."""
        count = 0
        ip = 0
        output = ""
        for ident in self.identifiers:
            count += 1
            for name in self.service_names:
                ip += 1
                networks = "net=10.0.0.{}".format(ip)
                if 'publisher' in name and ident.startswith('a'):
                    networks += ', 162.0.100.200'
                output += ('| id-{} | juju-{}-{}-machine-0 | {} |\n'.format(
                           count, name, ident, networks))
        return output

    def test_service_map_ignores_missing_identifiers(self):
        # Deployments with no identifiers on disk are ignored (they were
        # probably deployed manually with mojo.py, not with cd.py via cron).
        service_map = get_services_map(
            self.base_dir, nova_output=self.create_output())
        self.assertEqual({}, service_map)

    def test_service_map(self):
        self.generate_identifiers()
        service_map = get_services_map(
            self.base_dir, nova_output=self.create_output())

        # service_map is key-ed by services names.
        self.assertEqual(
            sorted(self.service_names), sorted(service_map.keys()))

        # Each service points to a dictionary key-ed by the deployment
        # identifier.
        service_contents = service_map['mojo-ue-core-image-watcher']
        self.assertEqual(
            ['56789', 'abcdef'], sorted(service_contents.keys()))

        # Each deployment contains 'path' (identifier file on disk), 'epoch'
        # (identifier file mtime) and 'units' (list of full unit names,
        # including the bootstrap node).
        deployment_contents = service_contents['abcdef']
        self.assertEqual(
            ['epoch', 'floating_ips', 'path', 'units'],
            sorted(deployment_contents.keys()))

    def test_service_map_floating_ips(self):
        self.generate_identifiers()
        service_map = get_services_map(
            self.base_dir, nova_output=self.create_output())
        self.assertEqual(service_map['mojo-ue-core-image-watcher']
                         ['abcdef']['floating_ips'], [])
        self.assertEqual(service_map['mojo-ue-core-image-publisher']
                         ['abcdef']['floating_ips'], ['162.0.100.200'])


if __name__ == '__main__':
    unittest.main()
