# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: miscelaneous utilities."""

import logging
import os
import subprocess
import textwrap


class BetterCalledProcessError(subprocess.CalledProcessError):

    """Extend CalledProcessError to show failed command output as well."""

    def __str__(self):
        return super().__str__() + self.output


def check_call(*popenargs, **kwargs):
    """A better version of subprocess.check_call that gives better errors.

    If a command invoked via 'check_call' fails, the error looks like::

        Command '['juju', 'bootstrap']' returned non-zero exit status 2

    While concise, that doesn't help the user of this script see what went
    wrong. This function invokes a command, but captures stdout and stderr,
    and includes that information in the exception.

    """
    if 'stdout' in kwargs or 'stderr' in kwargs:
        # user did something with stdout or stderr already, so just call
        # through to check_call directly:
        return subprocess.check_call(*popenargs, **kwargs)
    kwargs['stdout'] = kwargs['stderr'] = subprocess.PIPE
    process = subprocess.Popen(*popenargs, **kwargs)
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        raise BetterCalledProcessError(
            process.returncode,
            popenargs[0],
            textwrap.dedent(
                '''
                stdout:
                """\\
                {}"""
                stderr:
                """\\
                {}"""
                ''').format(stdout.decode(), stderr.decode())
        )


def setup_logger():
    """Return a configured logger instance.

    Defaults to INFO level and if 'CI_LOGSTASH_HOST' environment variable
    is set and python-logstash is reachable, adds a corresponding 'logstash'
    handler.
    """
    logging.basicConfig(
        format='%(asctime)s  %(name)s %(levelname)s: %(message)s')
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Optional remote logging via logstash.
    logstash_host = os.environ.get('CI_LOGSTASH_HOST')
    if logstash_host is not None:
        try:
            import logstash
        except ImportError:
            print('Follow the README instructions for installing '
                  'python-logstash')
            return 1
        else:
            logger.addHandler(
                logstash.LogstashHandler(logstash_host, 5959, 1))

    return logger
