#!/usr/bin/env python3
#
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""CI Automation: mojo spec deployer."""


import argparse
import configparser
import os
import sys


from ci_automation import (
    branch,
    config,
    juju,
    mojo,
    utils,
)


SERIES = 'trusty'


def fix_permissions(project):
    ''' "Why do I get permission denied?" from
    https://mojo.canonical.com/readme.html'''

    utils.check_call(['sudo', 'chmod', '755', '/var/lib/lxc'])
    path = '{}.{}'.format(project, SERIES)
    path = os.path.join('/var/lib/lxc', path)
    utils.check_call(['sudo', 'chmod', '755', path])


def create_service_conf(project, stage, **kwargs):
    c = configparser.ConfigParser()

    c['adt'] = {
        'max_retries': 3,
        'image_flavor': kwargs.get('adt_image_flavor'),
        'test_branch': 'lp:snappy/selftest',
        'archive_mirror': kwargs.get('adt_archive_mirror'),
        'http_proxy': kwargs.get('adt_http_proxy'),
        'os_username': kwargs.get('adt_os_username'),
        'os_password': kwargs.get('adt_os_password'),
        'os_tenant_name': kwargs.get('adt_os_tenant_name'),
        'os_auth_url': kwargs.get('adt_os_auth_url'),
        'os_region_name': kwargs.get('adt_os_region_name'),
        'extra_args': kwargs.get('adt_extra_args', ''),
    }

    c['image'] = {
        'channel': 'ubuntu-core/rolling/edge',
        'device': 'generic_amd64',
        'location': '/tmp/latest-core-image-version',
        'poll_period': '60'
    }

    c['swift'] = {
        'public_url': kwargs.get('swift_public_url') or '<not-set>',
    }

    c['image_mapper'] = {
        'url': kwargs.get('image_mapper_url') or '<not-set>',
    }

    c['logstash'] = {
        'host': kwargs.get('logstash_host') or '127.0.0.1',
        'port': 5959,
        'version': 1,
    }

    c['amqp'] = {
        'uris': kwargs.get(
            'amqp_uris') or 'amqp://guest:guest@localhost:5672//'
    }

    c['nova'] = {
        'os_username': kwargs.get('os_username') or os.environ['OS_USERNAME'],
        'os_tenant_name': kwargs.get(
            'os_tenant_name') or os.environ['OS_TENANT_NAME'],
        'os_password': kwargs.get('os_password') or os.environ['OS_PASSWORD'],
        'os_auth_url': kwargs.get('os_auth_url') or os.environ['OS_AUTH_URL'],
        'os_region_name': kwargs.get(
            'os_region_name') or os.environ['OS_REGION_NAME'],
    }

    # This is an ugly hack to keep things going for adt-cloud,
    # core-image and snappy-proposed solutions under the same tenant.
    if 'adt' in stage:
        filename = 'adt.conf'
    elif 'snappy-proposed' in stage:
        filename = 'snappy-proposed-service.conf'
    elif 'ols' in stage:
        filename = 'service.conf'
    else:
        filename = 'core-service.conf'

    path = '/srv/mojo/LOCAL/{}/{}/{}'.format(project, stage, filename)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as fp:
        c.write(fp)


def main():
    desc = 'Deploy a given mojo spec in isolation.'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('--project', default='mojo-stg-ue-ci-engineering')
    parser.add_argument('--stage', required=True)
    parser.add_argument('--branch', required=True)
    parser.add_argument('--devel', action="store_true", default=False)
    parser.add_argument('--base')
    parser.add_argument('--network')
    parser.add_argument('--amqp-uris')
    parser.add_argument('--logstash-host')
    parser.add_argument('--swift-public-url')
    parser.add_argument('--image-mapper-url')
    parser.add_argument('--os-username')
    parser.add_argument('--os-tenant-name')
    parser.add_argument('--os-password')
    parser.add_argument('--os-auth-url')
    parser.add_argument('--os-region-name')
    parser.add_argument('--adt-os-username', default='')
    parser.add_argument('--adt-os-tenant-name', default='')
    parser.add_argument('--adt-os-password', default='')
    parser.add_argument('--adt-os-auth-url', default='')
    parser.add_argument('--adt-os-region-name', default='')
    parser.add_argument('--adt-extra-args', default='')
    parser.add_argument('--adt-http-proxy', default='')
    args = parser.parse_args()

    base = args.base or '/srv/juju-environments'

    if not os.path.exists(base):
        print('Please create {}.'.format(base))
        return 1

    if not mojo.project_exists(args.project):
        msg = 'Please run `sudo mojo project-new --series trusty {}`.'
        msg = msg.format(args.project)
        print(msg)
        return 1

    # Le sigh.
    if args.devel:
        fix_permissions(args.project)

    # Thomi - we're moving to config files being stored on disk. As a
    # transitional stage, we use the new system if it has config on disk,
    # otherwise we fall back to the old system.
    if config.deployment_has_config(args.stage):
        # If config is on disk, warn the user if they're also specifying
        # config on the command line, as they might get confused otherwise.
        deprecated_args = (
            ('amqp_uris', None),
            ('logstash_host', None),
            ('swift_public_url', None),
            ('image_mapper_url', None),
            ('os_username', None),
            ('os_tenant_name', None),
            ('os_password', None),
            ('os_auth_url', None),
            ('os_region_name', None),
            ('adt_os_username', ''),
            ('adt_os_tenant_name', ''),
            ('adt_os_password', ''),
            ('adt_os_auth_url', ''),
            ('adt_os_region_name', ''),
            ('adt_extra_args', ''),
            ('adt_http_proxy', ''),
        )
        # argparse Namespace objects are kind of awkward to work with: Can't
        # find a better way to get namespace contents by varbale name:
        used_deprepcated = [
            a for a, d in deprecated_args
            if a in args and args.__getattribute__(a) != d
        ]
        if used_deprepcated:
            print("Warning: This mojo spec uses config on disk, but you have")
            print("specified one or more configuration options on the command")
            print("line. The following options will be ignored:")
            print(', '.join(used_deprepcated))
        config_target = os.path.join('/srv/mojo/LOCAL/', args.project)
        config.export_latest_config_for_deployment(args.stage, config_target)
    else:
        # Generate config from command line options:
        computed_adt_extra_args = ' '.join([
            args.adt_extra_args,
            '--net-id={}'.format(args.network) if args.devel else '',
        ]).strip()
        computed_adt_image_flavor = (
            'cpu2-ram2-disk50' if not args.devel else 'm1.small')
        computed_adt_archive_mirror = (
            'http://ftpmaster.internal/ubuntu/' if not args.devel else
            'http://nova.clouds.archive.ubuntu.com/ubuntu/')
        computed_adt_http_proxy = (
            'http://squid.internal:3128/' if not args.devel else
            args.adt_http_proxy
        )

        create_service_conf(
            args.project,
            args.stage,
            network=args.network,
            amqp_uris=args.amqp_uris,
            logstash_host=args.logstash_host,
            swift_public_url=args.swift_public_url,
            image_mapper_url=args.image_mapper_url,
            os_username=args.os_username,
            os_tenant_name=args.os_tenant_name,
            os_password=args.os_password,
            os_auth_url=args.os_auth_url,
            os_region_name=args.os_region_name,
            adt_os_username=args.adt_os_username,
            adt_os_tenant_name=args.adt_os_tenant_name,
            adt_os_password=args.adt_os_password,
            adt_os_auth_url=args.adt_os_auth_url,
            adt_os_region_name=args.adt_os_region_name,
            adt_extra_args=computed_adt_extra_args,
            adt_image_flavor=computed_adt_image_flavor,
            adt_archive_mirror=computed_adt_archive_mirror,
            adt_http_proxy=computed_adt_http_proxy,
        )

    # Relies on the canonical-mojo-specs layout, e.g.:
    # ue/mojo-ue-adt-cloud-worker/devel
    spec_name = args.stage.split('/')[1]

    # Juju environment & mojo workspaces have identical names. Names
    # are unique by appending a spec content identifier including
    # all collected branch URLs and their current revision numbers, e.g.:
    # mojo-ue-core-result-checker-c745311e948bd62221eafdcfcc552edf
    full_info_dict = branch.get_revnos_for_spec(args.branch, args.stage)
    identifier = branch.get_info_from_revno_dict(full_info_dict)
    env_name = '{}-{}'.format(spec_name, identifier)

    try:
        mojo.new_workspace(args.project, args.stage, args.branch, env_name)
        env_dir = juju.new_environment(
            env_name, base, devel=args.devel, network=args.network)
        mojo.deploy(args.project, args.stage, args.branch, env_dir)
        juju.set_metadata(env_name, base, full_info_dict)
    except utils.BetterCalledProcessError as err:
        print(err)
        return 1

    print('Deployed environment: {}'.format(env_dir))

    return 0


if __name__ == '__main__':
    sys.exit(main())
